# -*- test-case-name: exocet.test.test_exocet -*-
# Copyright (c) 2010 the authors. See LICENSE file for details.
import sys, __builtin__, itertools
from modules import getModule
from types import ModuleType

DEBUG = False


def trace(*args):
    if DEBUG:
        print ' '.join(str(x) for x in args)


def proxyModule(original, **replacements):
    """
    Create a proxy for a module object, overriding some of its attributes with
    replacement objects.

    @param original: A module.
    @param replacements: Attribute names and objects to associate with them.

    @returns: A module proxy with attributes containing the replacement
    objects; other attribute accesses are delegated to the original module.
    """
    class _ModuleProxy(object):
       def __getattribute__(self, name):
           if name in replacements:
               return replacements[name]
           else:
               return getattr(original, name)

       def __repr__(self):
           return "<Proxy for %r: %s replaced>" % (original, ', '.join(replacements.keys()))
    return _ModuleProxy()


def redirectLocalImports(name, globals=None, *a, **kw):
    """
    Catch function-level imports in modules loaded via Exocet. This ensures
    that any imports done after module load time look up imported names in the
    same context the module was originally loaded in.
    """

    if globals is not None:
        mf = globals.get('__exocet_context__', None)
        if mf is not None:
            trace("__import__ of", name,  "called in exocet module", mf, mf.mapper)
            return _isolateImports(mf, _originalImport, name, globals, *a, **kw)
        else:
            return _originalImport(name, globals, *a, **kw)
    else:
        return _originalImport(name, globals, *a, **kw)


_originalImport = None

def installExocetGlobalHook():
    """
    Install the global Exocet import hook.
    """
    global _originalImport
    _originalImport = __builtin__.__import__
    __builtin__.__import__ = redirectLocalImports



def uninstallExocetGlobalHook():
    __builtin__.__import__ = _originalImport


installExocetGlobalHook()

class Mapper(object):
    """
    An object that maps names used in C{import} statements to objects (such as
    modules).
    """
    def __init__(self, baseLookup):
        self._baseLookup = baseLookup


    def lookup(self, name):
        """
        Attempt to resolve a name to an object. Raise L{ImportError} if it's
        not present in this mapper.
        """
        try:
            return self._baseLookup(name)
        except ImportError:
            raise ImportError("No module named %r in mapper %r" % (name, self))


    def contains(self, name):
        """
        Return a boolean indicating whether this name can be resolved
        sucessfully by this mapper.
        """
        try:
            self.lookup(name)
            return True
        except ImportError:
            return False



class _PackageMapper(Mapper):
    """
    A mapper that allows direct mutation as a result of intrapackage
    imports. For internal use only. Void where prohibited. If symptoms
    persist, see a professional.
    """
    def __init__(self, name, baseMapper):
        self._name = name
        self._baseMapper = baseMapper
        self._intrapackageImports = {}


    def _baseLookup(self, name):
        if name in self._intrapackageImports:
            return self._intrapackageImports[name]
        else:
            return self._baseMapper.lookup(name)


    def add(self, name, module):
        """
        Add an intrapackage import.
        """
        #if not name.startswith(self._name):
        #    raise ValueError("%r is not a member of the %r package" %
        #                     (name, self._name))
        self._intrapackageImports[name] = module



def noLookup(name):
    raise ImportError(name)


emptyMapper = Mapper(noLookup)

class MakerFinder(object):
    """
    The object used as Exocet's PEP 302 meta-import hook. 'import' statements
    result in calls to find_module/load_module. A replacement for the
    C{__import__} function is provided as a method, as well.

    @ivar mapper: A L{Mapper}.

    @ivar oldImport: the implementation of C{__import__} being wrapped by this
                     object's C{xocImport} method.
    """
    def __init__(self, oldImport, mapper):
        self.mapper = mapper
        self.oldImport = oldImport


    def find_module(self, fullname, path=None):
        """
        Module finder nethod required by PEP 302 for meta-import hooks.

        @param fullname: The name of the module/package being imported.
        @param path: The __path__ attribute of the package, if applicable.
        """
        trace("find_module", fullname, path)
        return self


    def load_module(self, fqn):
        """
        Module loader method required by PEP 302 for meta-import hooks.

        @param fqn: The fully-qualified name of the module requested.
        """
        trace("load_module", fqn)
        trace("sys.modules", sys.modules)
        p = self.mapper.lookup(fqn)
        return p


    def xocImport(self, name, *args, **kwargs):
        """
        Wrapper around C{__import__}. Needed to ensure builtin modules aren't
        loaded from the global context.
        """
        trace("Import invoked:", name)
        if name in sys.builtin_module_names:
            trace("Loading builtin module", name)
            return self.load_module(name)
        else:
            return self.oldImport(name, *args, **kwargs)




def loadModule(fqn, mapper, m=None):
    """
    Load a Python module, eliminating as much of its access to global state as
    possible. If a package name is given, its __init__.py is loaded.

    @param fqn: The fully qualified name of a Python module, e.g
    C{twisted.python.filepath}.

    @param mapper: A L{Mapper}.

    @param m: An optional empty module object to load code into. (For
    resolving circular module imports.)

    @returns: An instance of the module name requested.
    """

    maker = getModule(fqn)
    if maker.isPackage() and m is None:
        pkgContext = PackageContext(maker, mapper, getModule)
        return pkgContext.loadModule(fqn)
    else:
        mf = MakerFinder(__builtin__.__import__, mapper)
        if maker.filePath.splitext()[1] in [".so", ".pyd"]:
            #it's native code, gotta suck it up and load it globally (really at a
            ## loss on how to unit test this without significant inconvenience)
            return maker.load()
        return _isolateImports(mf, _loadSingle, maker, mf, m)

def _loadSingle(mk, mf, m=None):
    trace("execfile", mk.name, m)
    if m is None:
        m = ModuleType(mk.name)
    contents = {}
    code = execfile(mk.filePath.path, contents)
    contents['__exocet_context__'] = mf
    m.__dict__.update(contents)
    m.__file__ = mk.filePath.path
    return m

def _isolateImports(mf, f, *a, **kw):
    """
    Internal guts for actual code loading. Displaces the global environment
    and executes the code, then restores the previous global settings.

    @param mk: A L{modules._modules.PythonModule} object; i.e., a module
    maker.
    @param mf: A L{MakerFinder} instance.
    @param m: An optional empty module object to load code into. (For resolving
    circular module imports.)
    """
    oldMetaPath = sys.meta_path
    oldPathHooks = sys.path_hooks
    oldSysModules = sys.modules.copy()
    oldImport = __builtin__.__import__
    #where is your god now?
    sys.path_hooks = ()
    sys.modules.clear()
    sys.meta_path = [mf]
    __builtins__['__import__'] = mf.xocImport

    #stupid special case for the stdlib
    if mf.mapper.contains('warnings'):
        sys.modules['warnings'] = mf.mapper.lookup('warnings')
    try:
       return f(*a, **kw)
    finally:
        sys.meta_path = oldMetaPath
        sys.path_hooks = oldPathHooks
        sys.modules.update(oldSysModules)
        __builtins__['__import__'] = oldImport

def loadPackage(fqn, mapper):
    """
    Analyze a Python package to determine its internal and external
    dependencies and produce an object capable of creating modules
    that refer to each other.

    @param fqn: The fully qualified name of a Python module, e.g
    C{twisted.python.filepath}.

    @param mapper: A L{Mapper}.

    @returns: A package loader object.
    """
    pkg = getModule(fqn)
    if not pkg.isPackage():
        raise ValueError("%r is not a package" % (fqn,))
    return PackageContext(pkg, mapper, getModule)


def _buildAndStoreEmptyModule(maker, mapper):
    m = ModuleType(maker.name)
    m.__path__ = maker.filePath.path
    mapper.add(maker.name,  m)
    return m


class PackageContext(object):
    """
    A context for loading interdependent modules from a package. This object
    allows modules in a package to refer to one another consistently without
    requiring them to be in sys.modules. Calling this object's C{loadModule}
    method will retrieve an already loaded instance if one exists, or else
    loads it (and all its previously unloaded dependencies within the package)
    and remembers them for later use.
    """
    def __init__(self, packageMaker, mapper, getModule):
        """
        @param packageMaker: A L{modules._modules.PythonModule} object
        representing a package.
        @param mapper: A L{Mapper}.
        @param getModule: A function to fetch module maker objects, given an
        FQN. (For example, {twisted.python.}modules.getModule.)
        """
        self.packageMaker = packageMaker
        self.mapper = _PackageMapper(packageMaker.name, mapper)
        initModule = _buildAndStoreEmptyModule(packageMaker, self.mapper)
        if '.' in packageMaker.name:
            modulePath = packageMaker.name.split(".")
            if not self.mapper.contains(modulePath[0]):
                parentMaker = getModule(modulePath[0])
                parent = _buildAndStoreEmptyModule(parentMaker, self.mapper)
                self.mapper.add(modulePath[0], parent)
                for seg in modulePath[1:-1]:
                    new = _buildAndStoreEmptyModule(parentMaker[seg], self.mapper)
                    setattr(parent, seg, new)
                    parent = new
            setattr(parent, modulePath[-1], initModule)
        self.allModules  = analyzePackage(self.packageMaker)
        sortedModules = robust_topological_sort(self.allModules)
        self.loadOrder = list(itertools.chain(*sortedModules))
        self.loadOrder.reverse()
        trace("load __init__", vars(initModule), id(initModule))
        self._reallyLoadModule(packageMaker.name, initModule)

    def _searchDeps(self, fqn, depsFoundSoFar=None):
        """
        Searches recursively for modules not yet loaded in this package.

        @param fqn: A module name.
        @param depsFoundSoFar: A set of dependencies found in previous
        invocations of this function, or None.
        """
        trace("searchDeps", fqn, depsFoundSoFar)
        if depsFoundSoFar is not None:
            deps = depsFoundSoFar
        else:
            deps = set()
        deps.add(fqn)
        maker, internalDeps, externalDeps = self.allModules[fqn]
        for dep in internalDeps:
            if dep not in deps and not self.mapper.contains(dep):
                    self._searchDeps(dep, deps)
        return deps

    def _unloadedDeps(self, fqn):
        """
        Returns an iterable of names of unloaded modules in this package that
        must be loaded to satisfy the dependencies of the named package, in
        the order they should be loaded.

        @param fqn: A module name in this package.
        """
        unloadedDeps = self._searchDeps(fqn)
        trace("raw unloaded deps", unloadedDeps)
        trace("load order", self.loadOrder)
        for dep in self.loadOrder:
            if dep in unloadedDeps:
                yield dep

    def loadModule(self, fqn):
        """
        Load the named module in this package, and any dependencies it may
        have, if not already loaded. Returns the named module.

        @param fqn: The fully-qualified name of a module in this package.
        """
        trace("package loadModule", fqn)
        if self.mapper.contains(fqn):
            trace("package module load satisfied from cache")
            return self.mapper.lookup(fqn)
        m = self._reallyLoadModule(fqn)
        trace("package module load triggered code loading", m)
        return m

    def _reallyLoadModule(self, fqn, m=None):
        """
        Do the actual work of loading modules (and their dependencies) into
        this package context.

        @param fqn: The fully-qualified name of a module in this package.
        @param m: An optional empty module object to load code into. (For resolving
        circular module imports.)
        """
        unloadedDeps = list(self._unloadedDeps(fqn))
        if self.packageMaker.name in unloadedDeps:
            unloadedDeps.remove(self.packageMaker.name)
        trace("unloaded deps", unloadedDeps)
        for name in unloadedDeps:
            self.mapper.add(name, ModuleType(name))
        for name in unloadedDeps:
            submod = self.mapper.lookup(name)
            loadModule(name, self.mapper, submod)
            trace("loaded", name, submod)
            setattr(self.mapper.lookup(self.packageMaker.name), name, submod)
        self.mapper.add(fqn, loadModule(fqn, self.mapper, m))
        return self.mapper.lookup(fqn)


def analyzePackage(pkg):
    """
    Build a graph of internal and external dependencies of all modules in a package.

    @param pkg: A L{modules._modules.PythonModule} object representing a
    package.

    @return: A mapping of fully qualified module names in a package to (module
    object, internal dependencies, external dependencies) triples.
    """
    allModules = {}
    def collect(module):
        if module.isPackage():
            for mod in module.walkModules():
                if mod is not module:
                    collect(mod)

        allModules[module.name] = (module, set(), set())


    collect(pkg)
    for name, (m, imports, externalDeps) in allModules.items():
        if m.filePath.splitext()[1] in [".so", ".pyd"]:
            continue
        for im in m.iterImportNames():
            prefix, dot, item = im.rpartition(".")
            if im in allModules:
                imports.add(im)
            elif prefix in allModules:
                imports.add(prefix)
            else:
                externalDeps.add(im)

    return allModules



## Topological sort based on Paul Harrison's public domain toposort.py

def strongly_connected_components(graph):
    """
    Find the strongly connected components in a graph using Tarjan's
    algorithm.

    @param graph: a mapping of node names to lists of successor nodes.
    """

    result = []
    stack = []
    low = {}

    def visit(node):
        if node in low: return

	num = len(low)
        low[node] = num
        stack_pos = len(stack)
        stack.append(node)

        for successor in graph[node][1]:
            visit(successor)
            low[node] = min(low[node], low[successor])

        if num == low[node]:
	    component = tuple(stack[stack_pos:])
            del stack[stack_pos:]
            result.append(component)
	    for item in component:
	        low[item] = len(graph)

    for node in graph:
        visit(node)

    return result


def topological_sort(graph):
    """
    Topological sort of a directed graph.

    @param graph: Mapping of nodes to lists of other nodes.

    @return: List of nodes, topologically sorted.
    """
    count = {}
    for node in graph:
        count[node] = 0
    for node in graph:
        for successor in graph[node]:
            count[successor] += 1

    ready = [node for node in graph if count[node] == 0]

    result = []
    while ready:
        node = ready.pop(-1)
        result.append(node)

        for successor in graph[node]:
            count[successor] -= 1
            if count[successor] == 0:
                ready.append(successor)

    return result


def robust_topological_sort(graph):
    """
    First identify strongly connected components,
    then perform a topological sort on these components.

    @param graph: Mapping of nodes to lists of other nodes.

    @return: List of tuples of strongly connected groups of nodes,
    topologically sorted.
    """
    components = strongly_connected_components(graph)

    node_component = {}
    for component in components:
        for node in component:
            node_component[node] = component

    component_graph = {}
    for component in components:
        component_graph[component] = []

    for node in graph:
        node_c = node_component[node]
        for successor in graph[node][1]:
            successor_c = node_component[successor]
            if node_c != successor_c:
                component_graph[node_c].append(successor_c)

    return topological_sort(component_graph)

