import sys
from unittest import TestCase
from exocet import loadPackage, loadModule, emptyMapper

class ModuleTests(TestCase):
    """
    Tests for loading individual modules.
    """

    def test_loadModule(self):
        """
        Modules can be loaded independent of global state.
        """
        m1 = loadModule("exocet.test.testpackage.util", emptyMapper)
        m1.testAttribute = "a value"

        m2 = loadModule("exocet.test.testpackage.util", emptyMapper)
        m2.testAttribute = "a different value"

        self.assertFalse(m1 is m2)
        self.assertNotEqual(m1.testAttribute, m2.testAttribute)
        self.assertFalse(m1 in sys.modules.values())
        self.assertFalse(m2 in sys.modules.values())


class MapperTests(TestCase):
    """
    Tests for objects that map names used in C{import} statements to module
    objects.
    """

    def test_emptyMapper(self):
        """
        The empty mapper rejects all lookups.
        """
        for name in ["sys", "email", "exocet", "exocet.test"]:
            self.assertRaises(ImportError, emptyMapper.lookup, name)
            self.assertFalse(emptyMapper.contains(name))



class PackageTests(TestCase):
    """
    Tests for package loading via Exocet.
    """

    def assertIdentical(self, left, right):
        """
        Assert two expressions refer to the same object.
        """
        self.assertTrue(left is right, "%s is not %s" % (repr(left), repr(right)))


    def test_loadPackage(self):
        """
        Calling C{loadPackage} returns a package context from which
        modules with shared state and knowledge of other code in the
        package can be loaded.
        """
        pc = loadPackage("exocet.test.testpackage", emptyMapper)
        top = pc.loadModule("exocet.test.testpackage.topmodule")
        foo = pc.loadModule("exocet.test.testpackage.foo")
        baz = pc.loadModule("exocet.test.testpackage.baz")
        util = pc.loadModule("exocet.test.testpackage.util")
        self.assertIdentical(top.foo, foo)
        self.assertIdentical(top.baz, baz)
        self.assertIdentical(foo.util, baz.util)


    def test_loadCircularPackage(self):
        """
        Packages with circular imports can be loaded correctly by C{loadPackage}.
        """

        pc = loadPackage("exocet.test.testpackage", emptyMapper)
        top = pc.loadModule("exocet.test.testpackage.topmodule")
        foo = pc.loadModule("exocet.test.testpackage.foo")
        baz = pc.loadModule("exocet.test.testpackage.baz")
        util = pc.loadModule("exocet.test.testpackage.util")
        self.assertIdentical(top.foo, foo)
        self.assertIdentical(top.baz, baz)
        self.assertIdentical(foo.util, baz.util)


    def test_loadPackageAsModule(self):
        """
        Calling L{loadModule} with the name of a package loads the package and returns its __init__ module.
        """

        m = loadModule("exocet.test.testpackage", emptyMapper)
        self.assertEqual(m.topName, "top value")
        self.assertEqual(m.fooName, "hooray")
        self.assertEqual(m.utilName, "hooray")




    def test_loadWithLocalImports(self):
        """
        Execution of local imports are resolved in the context the
        containing module was originally loaded in.
        """
        pc = loadPackage("exocet.test.testpackage_localimports", emptyMapper)
        foo = pc.loadModule("exocet.test.testpackage_localimports.foo")
        util = pc.loadModule("exocet.test.testpackage_localimports.util")
        self.assertEqual(foo.fooName, [])
        util2 = foo.do()
        self.assertEqual(foo.fooName[0], util.utilName)
        self.assertIdentical(util2, util)
